import logging
import os
import time
from pathlib import Path
from typing import Tuple

import numpy as np
import torch
from numpy._typing import NDArray
from pydantic import FilePath
from torch.utils.data.dataset import Dataset

from modalities.dataloader.open_gptx_dataset.mmap_dataset import make_dataset, print_rank_0


def get_num_tokens_per_epoch(documents: NDArray, sizes: NDArray) -> NDArray:
    """Total number of tokens in the dataset."""
    return np.sum(sizes[documents])


def get_num_epochs(num_tokens_per_epoch: int, sequence_len: int, num_samples: int) -> int:
    num_epochs = 0
    total_tokens = 0

    while True:
        num_epochs += 1
        total_tokens += num_tokens_per_epoch
        # -1 is because we need to retrieve seq_length + 1 token each time
        # but the last token will overlap with the first token of the next
        # sample except for the last sample.
        if ((total_tokens - 1) // sequence_len) >= num_samples:
            return num_epochs


def build_doc_idx(documents, num_epochs, np_rng, separate_last_epoch):
    """Build an array with length = number-of-epochs * number-of-documents.
    Each index is mapped to a corresponding document."""
    if not separate_last_epoch or num_epochs == 1:
        doc_idx = np.mgrid[0:num_epochs, 0 : len(documents)][1]
        doc_idx[:] = documents
        doc_idx = doc_idx.reshape(-1)
        doc_idx = doc_idx.astype(np.int32)
        np_rng.shuffle(doc_idx)

        return doc_idx

    doc_idx_first = build_doc_idx(documents, num_epochs - 1, np_rng, False)
    doc_idx_last = build_doc_idx(documents, 1, np_rng, False)

    return np.concatenate((doc_idx_first, doc_idx_last))


def _build_shuffle_idx(num_samples, total_size, np_rng):
    """Build the range [0, size) and shuffle."""
    print(
        " > building shuffle index with split [0, {}) and [{}, {}) " "...".format(num_samples, num_samples, total_size),
        flush=True,
    )

    dtype_ = np.uint32
    if total_size >= (np.iinfo(np.uint32).max - 1):
        dtype_ = np.int64

    shuffle_idx_first = np.arange(start=0, stop=num_samples, step=1, dtype=dtype_)
    np_rng.shuffle(shuffle_idx_first)
    if num_samples == total_size:
        return shuffle_idx_first

    shuffle_idx_last = np.arange(start=num_samples, stop=total_size, step=1, dtype=dtype_)
    np_rng.shuffle(shuffle_idx_last)

    return np.concatenate((shuffle_idx_first, shuffle_idx_last))


def build_sample_idx(
    sizes: NDArray,
    doc_idx: NDArray,
    seq_length: int,
    num_epochs: int,
    tokens_per_epoch: int,
):
    """Sample index mapping is a 2D array with sizes
    [number-of-samples + 1, 2] where [..., 0] contains
    the index into `doc_idx` and [..., 1] is the
    starting offset in that document."""

    # Total number of samples. For -1 see comments in `_num_epochs`.
    num_samples = (num_epochs * tokens_per_epoch - 1) // seq_length
    sample_idx = np.zeros([num_samples + 1, 2], dtype=np.int32)

    # Index into sample_idx.
    sample_index = 0
    # Index into doc_idx.
    doc_idx_index = 0
    # Begining offset for each document.
    doc_offset = 0
    # Start with first document and no offset.
    sample_idx[sample_index][0] = doc_idx_index
    sample_idx[sample_index][1] = doc_offset
    sample_index += 1
    while sample_index <= num_samples:
        # Start with a fresh sequence.
        remaining_seq_length = seq_length + 1
        while remaining_seq_length != 0:
            # Get the document length.
            doc_id = doc_idx[doc_idx_index]
            doc_length = sizes[doc_id] - doc_offset
            # And add it to the current sequence.
            remaining_seq_length -= doc_length
            # If we have more than a full sequence, adjust offset and set
            # remaining length to zero so we return from the while loop.
            # Note that -1 here is for the same reason we have -1 in
            # `_num_epochs` calculations.
            if remaining_seq_length <= 0:
                doc_offset += remaining_seq_length + doc_length - 1
                remaining_seq_length = 0
            else:
                # Otherwise, start from the begining of the next document.
                doc_idx_index += 1
                doc_offset = 0
        # Record the sequence.
        sample_idx[sample_index][0] = doc_idx_index
        sample_idx[sample_index][1] = doc_offset
        sample_index += 1

    return sample_idx


def build_index_mappings(
    name: str,
    data_prefix: str,
    documents: NDArray,
    sizes: NDArray,
    num_samples: int,
    sequence_len: int,
    seed: int,
    cutoff_ratio_last_epoch: float = 0.95,
    is_distributed: bool = True,
) -> Tuple[NDArray, NDArray, NDArray]:
    """Build doc-idx, sample-idx, and shuffle-idx.
    doc-idx: is an array (ordered) of documents to be used in training.
    sample-idx: is the start document index and document offset for each training sample.
    shuffle-idx: maps the sample index into a random index into sample-idx.
    :param name: 'train', 'validation' or 'test'
    :type name:
    :param data_prefix:
    :type data_prefix:
    :param documents:
    :type documents:
    :param seed:
    :type seed:
    :param cutoff_ratio_last_epoch:
    :type cutoff_ratio_last_epoch:
    :return:
    :rtype:
    """
    # Number of tokens in each epoch and number of required epochs.
    tokens_per_epoch = get_num_tokens_per_epoch(sizes=sizes, documents=documents)
    num_epochs = get_num_epochs(
        num_tokens_per_epoch=tokens_per_epoch,
        sequence_len=sequence_len,
        num_samples=num_samples,
    )

    # rng state
    np_rng = np.random.RandomState(seed=seed)

    # Filename of the index mappings.
    _filename = data_prefix
    _filename += "_{}_indexmap".format(name)
    _filename += "_{}ns".format(num_samples)
    _filename += "_{}sl".format(sequence_len)
    _filename += "_{}s".format(seed)
    doc_idx_filename = f"{_filename}_doc_idx.npy"
    sample_idx_filename = f"{_filename}_sample_idx.npy"
    shuffle_idx_filename = f"{_filename}_shuffle_idx.npy"

    # Build the indexed mapping if not exist.
    if not is_distributed or torch.distributed.get_rank() == 0:
        if (
            (not os.path.isfile(doc_idx_filename))
            or (not os.path.isfile(sample_idx_filename))
            or (not os.path.isfile(shuffle_idx_filename))
        ):
            print_rank_0(" > WARNING: could not find index map files, building the indices on rank 0 ...")

            # For the last epoch, decide whether include the entire epoch
            # in the global shuffle or not.

            # If we need only one epoch, then separating last epoch  does
            # not mean anything.
            if num_epochs == 1:
                separate_last_epoch = False
                print(
                    " > only one epoch required, setting " "separate_last_epoch to False",
                    flush=True,
                )

            else:
                # Get the number of samples for the last epoch
                num_samples_from_epochs_minus_one = ((num_epochs - 1) * tokens_per_epoch - 1) // sequence_len
                last_epoch_num_samples = num_samples - num_samples_from_epochs_minus_one

                assert (
                    last_epoch_num_samples >= 0
                ), f"last epoch number of samples {last_epoch_num_samples} should be non-negative."
                num_samples_per_epoch = (tokens_per_epoch - 1) // sequence_len
                assert (
                    last_epoch_num_samples <= num_samples_per_epoch
                ), f"last epoch number of samples {last_epoch_num_samples} exceeded max value {num_samples_per_epoch}."
                # If we have less than cutoff_last_epoch * samples_per_epoch of the samples for the last epoch,
                # seperate out the epoch and treat it differently.
                separate_last_epoch = last_epoch_num_samples < int(cutoff_ratio_last_epoch * num_samples_per_epoch)
                if separate_last_epoch:
                    string = (
                        " > last epoch number of samples ({}) is smaller "
                        "than {}% of number of samples per epoch ({}), "
                        "setting separate_last_epoch to True"
                    )
                else:
                    string = (
                        " > last epoch number of samples ({}) is larger "
                        "than {}% of number of samples per epoch ({}), "
                        "setting separate_last_epoch to False"
                    )
                print(
                    string.format(last_epoch_num_samples, cutoff_ratio_last_epoch * 100, num_samples_per_epoch),
                    flush=True,
                )

            # doc-idx.
            start_time = time.time()
            doc_idx = build_doc_idx(
                documents,
                num_epochs,
                np_rng,
                separate_last_epoch,
            )
            np.save(doc_idx_filename, doc_idx, allow_pickle=True)
            print_rank_0(
                " > elasped time to build and save doc-idx mapping " "(seconds): {:4f}".format(time.time() - start_time)
            )
            # sample-idx.
            start_time = time.time()
            # Use C++ implementation for speed.
            # First compile and then import.
            # TODO: check import

            assert doc_idx.dtype == np.int32
            assert sizes.dtype == np.int32
            # from rlm.util.helpers import helpers
            sample_idx = build_sample_idx(
                sizes=sizes,
                doc_idx=doc_idx,
                seq_length=sequence_len,
                num_epochs=num_epochs,
                tokens_per_epoch=tokens_per_epoch,
            )

            np.save(sample_idx_filename, sample_idx, allow_pickle=True)
            print_rank_0(
                " > elasped time to build and save sample-idx mapping "
                "(seconds): {:4f}".format(time.time() - start_time)
            )
            # shuffle-idx.
            start_time = time.time()
            # -1 is due to data structure used to retieve the index:
            #    sample i --> [sample_idx[i], sample_idx[i+1])
            if separate_last_epoch:
                num_samples_ = num_samples_from_epochs_minus_one
            else:
                num_samples_ = sample_idx.shape[0] - 1
            shuffle_idx = _build_shuffle_idx(
                num_samples_,
                sample_idx.shape[0] - 1,
                np_rng,
            )
            np.save(shuffle_idx_filename, shuffle_idx, allow_pickle=True)
            print_rank_0(
                " > elasped time to build and save shuffle-idx mapping"
                " (seconds): {:4f}".format(time.time() - start_time)
            )

    # This should be a barrier but nccl barrier assumes
    # device_index=rank which is not the case for model
    # parallel case
    # counts = torch.cuda.LongTensor([1])

    # if is_distributed
    #   torch.distributed.all_reduce(counts, group=mpu.get_data_parallel_group())
    #   torch.distributed.all_reduce(counts, group=mpu.get_pipeline_model_parallel_group())

    # TODO: check: in MegatronLM, an all_reduce() (see below) is called because of the different model parallel groups,
    #  i.e., not all ranks are in the data parallel group
    if is_distributed:
        torch.distributed.barrier()

    # if is_distributed
    #   assert counts[0].item() == (
    #   torch.distributed.get_world_size() //
    #   torch.distributed.get_world_size(group=mpu.get_tensor_model_parallel_group()))

    # Load mappings.
    start_time = time.time()
    print_rank_0(" > loading doc-idx mapping from {}".format(doc_idx_filename))
    doc_idx = np.load(doc_idx_filename, allow_pickle=True, mmap_mode="r")
    print_rank_0(" > loading sample-idx mapping from {}".format(sample_idx_filename))
    sample_idx = np.load(sample_idx_filename, allow_pickle=True, mmap_mode="r")
    print_rank_0(" > loading shuffle-idx mapping from {}".format(shuffle_idx_filename))
    shuffle_idx = np.load(shuffle_idx_filename, allow_pickle=True, mmap_mode="r")
    print_rank_0("    loaded indexed file in {:3.3f} seconds".format(time.time() - start_time))
    print_rank_0("    total number of samples: {}".format(sample_idx.shape[0]))
    print_rank_0("    total number of epochs: {}".format(num_epochs))

    return doc_idx, sample_idx, shuffle_idx


class OpenGPTXMMapDataset(Dataset):
    def __init__(
        self,
        sample_key: str,
        path: FilePath,
        # text_dataset: MMapIndexedDataset,
        # dataset_dir: str,
        # doc_idx: NDArray[int],
        # dataset_name: str,
        sequence_len: int,
        num_samples: int,
        seed: int = 47,
    ):
        """
        :param text_dataset:
        :type text_dataset:
        :param dataset_dir:
        :type dataset_dir:
        :param doc_idx: document ids. A document can contain several samples.
        :type doc_idx:
        :param dataset_name: train, validation or test.
        :type dataset_name:
        :param sequence_len:
        :type sequence_len:
        :param num_samples: number of samples contained in the training/validation/test set.
        :type num_samples:
        :param seed:
        :type seed:

        """

        dataset_dir = path.parents[0]
        dataset_filename_prefix = path.stem
        text_dataset = make_dataset(path=dataset_dir.joinpath(dataset_filename_prefix))

        doc_idx = np.arange(0, len(text_dataset))

        self.sample_key = sample_key
        # Contains the document inidices
        self.doc_idx = doc_idx
        self.text_dataset = text_dataset
        # + 1, because we shift sequence in order to get source and target sequence
        self.sequence_len = sequence_len
        self.num_samples = num_samples

        logging.info("Compiling dataset index builder.")

        (
            self.doc_idx,
            self.sample_idx,
            self.shuffle_idx,
        ) = build_index_mappings(
            name=dataset_filename_prefix,
            sizes=text_dataset.sizes,
            num_samples=num_samples,
            sequence_len=self.sequence_len,
            data_prefix=Path(dataset_dir).joinpath(text_dataset._path).__str__(),
            # as input for the next indexed dataset, where it gets shuffled again. Is this correct?
            documents=self.doc_idx,
            seed=seed,
        )

    def __len__(self):
        # -1 is due to data structure used to retieve the index:
        #    sample i --> [sample_idx[i], sample_idx[i+1])
        return self.sample_idx.shape[0] - 1

    def __getitem__(self, idx: int):
        # Get the shuffled index.
        idx = self.shuffle_idx[idx]
        # Start and end documents and offsets.
        doc_index_f = self.sample_idx[idx][0]
        doc_index_l = self.sample_idx[idx + 1][0]

        offset_f = self.sample_idx[idx][1]
        offset_l = self.sample_idx[idx + 1][1]
        # If we are within the same document, just extract the chunk.
        doc_ids = []
        if doc_index_f == doc_index_l:
            doc_ids.append(self.doc_idx[doc_index_f])
            sample = self.text_dataset.get(
                self.doc_idx[doc_index_f],
                offset=offset_f,
                length=offset_l - offset_f + 1,
            )
        else:
            # Otherwise, get the rest of the initial document.
            doc_ids.append(self.doc_idx[doc_index_f])
            sample_list = [self.text_dataset.get(self.doc_idx[doc_index_f], offset=offset_f)]

            # Loop over all in between documents and add the entire document.
            for i in range(doc_index_f + 1, doc_index_l):
                doc_ids.append(self.doc_idx[i])
                sample_list.append(self.text_dataset.get(self.doc_idx[i]))

            # And finally add the relevant portion of last document.
            doc_ids.append(self.doc_idx[doc_index_l])
            sample_list.append(self.text_dataset.get(self.doc_idx[doc_index_l], length=offset_l + 1))
            sample = np.concatenate(sample_list)

        # Sample is of length sequence_len + 1 because target toke is part of the sample
        return {self.sample_key: np.array(sample, dtype=np.int64)}


class OpenGPTXMMapDatasetFactory:
    @staticmethod
    def create_dataset(num_samples: int, path: FilePath, sample_key: str, sequence_len: int) -> OpenGPTXMMapDataset:
        # dataset_dir = path.parents[0]
        # dataset_filename_prefix = path.stem
        # text_dataset = make_dataset(path=dataset_dir.joinpath(dataset_filename_prefix))

        # instances = OpenGPTXDataset(
        #     sample_key=sample_key,
        #     text_dataset=text_dataset,
        #     doc_idx=np.arange(0, len(text_dataset)),
        #     dataset_dir=dataset_dir,
        #     num_samples=num_samples,
        #     dataset_name=dataset_filename_prefix,
        #     sequence_len=sequence_len,
        # )
        # return instances
        pass
